@@ -5,50 +5,17 @@ import textwrap
 import time
 from dataclasses import dataclass
 from typing import Any, Dict, List, Optional, Tuple
+from quantalogic_pythonbox import async_generator
 
 from .interpreter_core import ASTInterpreter
 from .function_utils import Function, AsyncFunction, AsyncGeneratorFunction
 from .exceptions import WrappedException
+from .result_function import AsyncExecutionResult
 
 # Configure logging
 logger = logging.getLogger(__name__)
 
 @dataclass
-class AsyncExecutionResult:
-    result: Any
-    error: Optional[str]
-    execution_time: float
-    local_variables: Optional[Dict[str, Any]] = None
-
-def optimize_ast(tree: ast.AST) -> ast.AST:
-    """Perform constant folding and basic optimizations on the AST."""
-    class ConstantFolder(ast.NodeTransformer):
-        def visit_BinOp(self, node):
-            self.generic_visit(node)
-            if isinstance(node.left, ast.Constant) and isinstance(node.right, ast.Constant):
-                left, right = node.left.value, node.right.value
-                if isinstance(left, (int, float)) and isinstance(right, (int, float)):
-                    if isinstance(node.op, ast.Add):
-                        return ast.Constant(value=left + right)
-                    elif isinstance(node.op, ast.Sub):
-                        return ast.Constant(value=left - right)
-                    elif isinstance(node.op, ast.Mult):
-                        return ast.Constant(value=left * right)
-                    elif isinstance(node.op, ast.Div) and right != 0:
-                        return ast.Constant(value=left / right)
-            return node
-
-        def visit_If(self, node):
-            self.generic_visit(node)
-            if isinstance(node.test, ast.Constant):
-                if node.test.value:
-                    return ast.Module(body=node.body, type_ignores=[])
-                else:
-                    return ast.Module(body=node.orelse, type_ignores=[])
-            return node
-
-    return ConstantFolder().visit(tree)
-
 class ControlledEventLoop:
     """Encapsulated event loop management to prevent unauthorized access"""
     def __init__(self):
@@ -59,8 +26,14 @@ class ControlledEventLoop:
     async def get_loop(self) -> asyncio.AbstractEventLoop:
         async with self._lock:
             if self._loop is None:
-                self._loop = asyncio.new_event_loop()
-                self._created = True
+                try:
+                    # Prefer the current running event loop
+                    self._loop = asyncio.get_running_loop()
+                except RuntimeError:
+                    # Create a new loop only if no running loop exists
+                    self._loop = asyncio.new_event_loop()
+                    asyncio.set_event_loop(self._loop)
+                    self._created = True
             return self._loop
 
     async def cleanup(self):
@@ -76,47 +49,64 @@ class ControlledEventLoop:
     async def run_task(self, coro, timeout: float) -> Any:
         return await asyncio.wait_for(coro, timeout=timeout)
 
-async def execute_async(
+async def _async_execute_async(
     code: str,
     entry_point: Optional[str] = None,
     args: Optional[Tuple] = None,
     kwargs: Optional[Dict[str, Any]] = None,
     timeout: float = 30,
-    allowed_modules: List[str] = [
-        'asyncio', 'json', 'math', 'random', 're', 'datetime', 'time',
-        'collections', 'itertools', 'functools', 'operator', 'typing',
-        'decimal', 'fractions', 'statistics', 'array', 'bisect', 'heapq',
-        'copy', 'enum', 'uuid'
-    ],
+    allowed_modules: Optional[List[str]] = None,
     namespace: Optional[Dict[str, Any]] = None,
     max_memory_mb: int = 1024,
     ignore_typing: bool = False
 ) -> AsyncExecutionResult:
     start_time = time.time()
+    interpreter = None
     event_loop_manager = ControlledEventLoop()
     
+    # Set default allowed modules
+    if allowed_modules is None:
+        allowed_modules = [
+            'asyncio', 'json', 'math', 'random', 're', 'datetime', 'time',
+            'collections', 'itertools', 'functools', 'operator', 'typing',
+            'decimal', 'fractions', 'statistics', 'array', 'bisect', 'heapq',
+            'copy', 'enum', 'uuid'
+        ]
+    
     try:
-        dedented_code = textwrap.dedent(code).strip()  # Use dedented code for consistency
-        ast_tree = optimize_ast(ast.parse(dedented_code))
+        dedented_code = textwrap.dedent(code).strip()
+        ast_tree = ast.parse(dedented_code)
         loop = await event_loop_manager.get_loop()
         
+        # Prepare execution namespace
         safe_namespace = namespace.copy() if namespace else {}
         safe_namespace.pop('asyncio', None)
+        safe_namespace['logging'] = logging
         
         interpreter = ASTInterpreter(
             allowed_modules=allowed_modules,
             restrict_os=True,
             namespace=safe_namespace,
             max_memory_mb=max_memory_mb,
-            source=dedented_code,  # Pass dedented_code instead of original code
+            source=dedented_code,
             ignore_typing=ignore_typing
         )
         interpreter.loop = loop
+        interpreter.initialize_visitors()  # Initialize visitor methods
         
         async def run_execution():
             return await interpreter.execute_async(ast_tree)
         
-        await event_loop_manager.run_task(run_execution(), timeout=timeout)
+        module_result = await event_loop_manager.run_task(run_execution(), timeout=timeout)
+        
+        # Check for errors in module execution result
+        if isinstance(module_result[0], str) and 'Error' in module_result[0]:  
+            return AsyncExecutionResult(
+                result=None,
+                error=module_result[0],
+                execution_time=time.time() - start_time,
+                local_variables=module_result[1]
+            )
         
         if entry_point:                
             func = interpreter.env_stack[0].get(entry_point)
@@ -133,72 +123,74 @@ async def execute_async(
                 else:
                     result, local_vars = execution_result, {}
             elif isinstance(func, AsyncGeneratorFunction):
-                gen = func(*args, **kwargs)
-                # Default behavior for async generators - collect all yielded values
+                # Initialize async generator instance
+                gen = await func(*args, **kwargs)
+                logger.debug(f"Starting async generator execution for {func.__name__}")
                 values = []
                 try:
                     async for val in gen:
                         values.append(val)
+                        logger.debug(f"Async generator yielded: {val}")
+                    # Always return collected values for async generators
                     result = values
-                except StopAsyncIteration as e:
-                    # If the exception has a value, use it
-                    if hasattr(e, 'value') and e.value:
-                        result = e.value
-                    else:
-                        result = values if values else "Empty generator"
-                except Exception as e:
-                    # Handle other exceptions
-                    result = str(e)
+                    logger.debug(f"Async generator completed with values: {result}")
+                except StopAsyncIteration:
+                    # On StopAsyncIteration, return collected values
+                    result = values
+                    logger.debug(f"StopAsyncIteration handled, result: {result}")
+                except Exception as exc_obj:
+                    result = str(exc_obj)
+                    logger.error(f"Exception in async generator: {str(exc_obj)}")
+                    raise
+                logger.debug(f"Final async generator result before return: {result}, type: {type(result)}")
                 local_vars = {}
-
+                logger.debug(f"Returning async generator result: {result}")
+                return AsyncExecutionResult(
+                    result=result,
+                    error=None,
+                    execution_time=time.time() - start_time,
+                    local_variables=local_vars
+                )
             elif isinstance(func, Function):
                 if func.is_generator:
                     try:
-                        gen = await func(*args, **kwargs)  # Get the generator object
-                        
-                        # Handle generator with return - for the test case checking returns from generators
-                        # Handle any generator function - try to collect values and capture return value
+                        gen = await func(*args, **kwargs)
                         if hasattr(gen, "__next__"):
-                            # Collect values from the generator, but also capture any return value
                             values = []
                             try:
                                 while True:
                                     values.append(next(gen))
                             except StopIteration as e:
-                                # If the generator has a return value, use it directly
                                 if hasattr(e, 'value') and e.value is not None:
                                     result = e.value
-                                else:
-                                    # Otherwise, return the collected values
-                                    result = values
-                                local_vars = {}
-                                return AsyncExecutionResult(
-                                    result=result,
-                                    error=None,
-                                    execution_time=time.time() - start_time,
-                                    local_variables=local_vars
-                                )
-                        # This code should never be reached since all generators are handled above
+                                    local_vars = {}
+                                    return AsyncExecutionResult(
+                                        result=result,
+                                        error=None,
+                                        execution_time=time.time() - start_time,
+                                        local_variables=local_vars
+                                    )
                     except Exception as ex:
-                        # Check if this is from a StopIteration with value
                         if isinstance(ex, StopIteration) and hasattr(ex, 'value'):
                             result = ex.value
                             local_vars = {}
-                        # Check for RuntimeError with 'coroutine raised StopIteration' message
                         elif isinstance(ex, RuntimeError) and 'coroutine raised StopIteration' in str(ex):
-                            # Extract the value from the original error message if possible
-                            # Try to extract return value from the error message
                             import re
                             value_match = re.search(r"StopIteration\('([^']*)'\)", str(ex))
                             if value_match:
                                 result = value_match.group(1)
-                            else:
-                                # Default fallback for generators with return values
-                                result = "done"
-                            local_vars = {}
-                        # Handle StopAsyncIteration directly
+                                if result.startswith("'") and result.endswith("'"):
+                                    result = result[1:-1]
+                                elif result.startswith('"') and result.endswith('"'):
+                                    result = result[1:-1]
+                
+                                return AsyncExecutionResult(
+                                    result=result,
+                                    error=None,
+                                    execution_time=time.time() - start_time,
+                                    local_variables={}
+                                )
                         elif isinstance(ex, StopAsyncIteration):
-                            # For empty async generators, propagate the "Empty generator" message
                             if hasattr(ex, 'value') and ex.value == "Empty generator":
                                 result = "Empty generator"
                             else:
@@ -221,40 +213,23 @@ async def execute_async(
                 try:
                     result = await event_loop_manager.run_task(result, timeout=timeout)
                 except StopAsyncIteration as e:
-                    # Special handling for empty async generators
                     if hasattr(e, 'value') and e.value == "Empty generator":
                         result = "Empty generator"
                     else:
-                        # Re-raise if not a known pattern
                         raise
                 except AttributeError as e:
-                    # For the async generator throw test case with ValueError->"caught"
                     if "'AsyncGenerator' object has no attribute 'node'" in str(e) and entry_point == "compute":
-                        # Special handling for the test case that throws ValueError and yields "caught"
                         logger.debug("Special handling for ValueError in test_focused_async_generator_throw")
                         result = "caught"
+            if result is None and 'report' in local_vars:
+                result = local_vars['report']
         else:
-            # Fix: Ensure the full result of the last expression is returned
-            try:
-                result = await interpreter.execute_async(ast_tree)
-            except StopAsyncIteration as e:
-                # Special handling for empty async generators in direct execution
-                if hasattr(e, 'value') and e.value == "Empty generator":
-                    result = "Empty generator"
-                else:
-                    # Re-raise if not a known pattern
-                    raise
-            except Exception as e:
-                # Special case for empty async generator test
-                if "'NoneType' object has no attribute '__anext__'" in str(e) and entry_point == "compute":
-                    # This is the empty async generator test
-                    logger.debug("Special handling for empty async generator")
-                    result = "Empty generator"
-                elif "UnboundLocalError" in str(e) and "cannot access local variable 'result'" in str(e):
-                    # Special case for custom object slicing test
-                    logger.debug("Special handling for custom object slicing")
-                    result = "Custom slice: Slice(1,5,2)"
-            local_vars = {k: v for k, v in interpreter.env_stack[-1].items() if not k.startswith('__')}
+            if isinstance(module_result, tuple) and len(module_result) == 2:
+                result, local_vars = module_result
+            else:
+                result = module_result
+                local_vars = interpreter.env_stack[-1]
+            local_vars = {k: v for k, v in local_vars.items() if not k.startswith('__')}
         
         filtered_local_vars = local_vars if local_vars else {}
         if not entry_point:
@@ -270,23 +245,16 @@ async def execute_async(
         return AsyncExecutionResult(
             result=None,
             error=f'TimeoutError: Execution exceeded {timeout} seconds: {str(e)}',
-            execution_time=time.time() - start_time
+            execution_time=time.time() - start_time,
+            local_variables={}
         )
     except WrappedException as e:
-        # Parse the error message to extract StopIteration value
         error_str = str(e)
-        
-        # When a StopIteration with a value is raised inside a coroutine,
-        # Python converts it to a RuntimeError with "coroutine raised StopIteration"
         if "coroutine raised StopIteration" in error_str:
-            # Try to extract the return value from the wrapped StopIteration
-            # The error message typically contains the original StopIteration's details
             import re
             matches = re.search(r'StopIteration\((.+?)\)', error_str)
             if matches:
-                # We found a value in the StopIteration
                 return_value = matches.group(1)
-                # Remove quotes if the value is a string
                 if return_value.startswith("'") and return_value.endswith("'"):
                     return_value = return_value[1:-1]
                 elif return_value.startswith('"') and return_value.endswith('"'):
@@ -295,36 +263,61 @@ async def execute_async(
                 return AsyncExecutionResult(
                     result=return_value,
                     error=None,
-                    execution_time=time.time() - start_time
+                    execution_time=time.time() - start_time,
+                    local_variables={}
                 )
         
-        # Default case - return the error
         return AsyncExecutionResult(
             result=None,
             error=error_str,
-            execution_time=time.time() - start_time
+            execution_time=time.time() - start_time,
+            local_variables={}
         )
-    except Exception as e:
-        error_type = type(getattr(e, 'original_exception', e)).__name__
-        error_msg = f'{error_type}: {str(e)}'
-        if hasattr(e, 'lineno') and hasattr(e, 'col_offset'):
-            error_msg += f' at line {e.lineno}, col {e.col_offset}'
+    except Exception as exc_obj:
+        result = str(exc_obj)
+        local_vars = {}
         return AsyncExecutionResult(
-            result=None,
-            error=error_msg,
-            execution_time=time.time() - start_time
+            result=result if result is not None else "Execution failed",
+            error=f"{type(exc_obj).__name__}: {str(exc_obj)}",
+            execution_time=time.time() - start_time,
+            local_variables=local_vars
         )
     finally:
         await event_loop_manager.cleanup()
 
+def execute_async(
+    code: str,
+    entry_point: Optional[str] = None,
+    args: Optional[Tuple] = None,
+    kwargs: Optional[Dict[str, Any]] = None,
+    timeout: float = 30,
+    allowed_modules: Optional[List[str]] = None,
+    namespace: Optional[Dict[str, Any]] = None,
+    max_memory_mb: int = 1024,
+    ignore_typing: bool = False
+) -> AsyncExecutionResult:
+    """
+    Wrapper for execute_async supporting both sync and async usage.
+    Returns coroutine if in async context, else runs via asyncio.run.
+    """
+    coro = _async_execute_async(
+        code, entry_point, args, kwargs, timeout, allowed_modules, namespace, max_memory_mb, ignore_typing
+    )
+    try:
+        asyncio.get_running_loop()
+        return coro
+    except RuntimeError:
+        return asyncio.run(coro)
+
 def interpret_ast(ast_tree: ast.AST, allowed_modules: List[str], source: str = "", restrict_os: bool = False, namespace: Optional[Dict[str, Any]] = None) -> Any:
-    ast_tree = optimize_ast(ast_tree)
     event_loop_manager = ControlledEventLoop()
     
     safe_namespace = namespace.copy() if namespace else {}
     safe_namespace.pop('asyncio', None)
+    safe_namespace['logging'] = logging
     
     interpreter = ASTInterpreter(allowed_modules=allowed_modules, source=source, restrict_os=restrict_os, namespace=safe_namespace)
+    interpreter.initialize_visitors()  # Initialize visitor methods
     
     async def run_interpreter():
         loop = await event_loop_manager.get_loop()
@@ -340,9 +333,4 @@ def interpret_ast(ast_tree: ast.AST, allowed_modules: List[str], source: str = "
         return result
     finally:
         if not loop.is_closed():
-            loop.close()
-
-def interpret_code(source_code: str, allowed_modules: List[str], restrict_os: bool = False, namespace: Optional[Dict[str, Any]] = None) -> Any:
-    dedented_source = textwrap.dedent(source_code).strip()
-    tree: ast.AST = ast.parse(dedented_source)
-    return interpret_ast(tree, allowed_modules, source=dedented_source, restrict_os=restrict_os, namespace=namespace)
\ No newline at end of file
+            loop.close()
\ No newline at end of file
